
#ifdef _WIN32

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "drvmgr.h"

static UCHAR driverLocation[MAX_PATH];
static HANDLE fileHandle;

HANDLE getHandle(void) {
    return fileHandle;
}

int LoadDriver(void) {
    UCHAR serviceName[MAX_PATH];
    DWORD errNum = 0;

    // Set up the service name.

    strcpy(serviceName, "\\\\.\\"DRIVER_NAME);

    // Try to connect to driver.  If this fails, try to load the driver
    // dynamically.
    if ((fileHandle = CreateFile(serviceName, GENERIC_READ | GENERIC_WRITE,
          0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE) {
        errNum = GetLastError();
        if (errNum != ERROR_FILE_NOT_FOUND) {
            printf("CreateFile failed!  Error = %d\n", errNum);
            return -1;
        }

        // Setup full path to driver name.
        if (!SetupDriverName(driverLocation)) {
            return -1;
        }

        // Install driver.
        if (!ManageDriver(DRIVER_NAME, driverLocation, DRIVER_FUNC_INSTALL)) {
            printf("Unable to install driver. \n");
            // Error - remove driver.
            ManageDriver(DRIVER_NAME, driverLocation, DRIVER_FUNC_REMOVE);
            CloseHandle(fileHandle);
            return -1;
        }

        // Try to open the newly installed driver.
        if ((fileHandle = CreateFile(serviceName, GENERIC_READ | GENERIC_WRITE,
             0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL)) == INVALID_HANDLE_VALUE) {
            printf("CreateFile failed!  Error = %d\n", GetLastError() );
            return -1;
        }
    } else {
       printf("=== LoadDriver succesfull.\n");
    }
    return 0;
}

void UnloadDriver(void) {
    if (fileHandle) {
        CloseHandle(fileHandle);
    }

    // Unload the driver.  Ignore any errors.
    ManageDriver(DRIVER_NAME, driverLocation, DRIVER_FUNC_REMOVE);
    printf("=== UnloadDriver.\n");
}

BOOLEAN InstallDriver(SC_HANDLE SchSCManager, LPCTSTR DriverName, LPCTSTR ServiceExe) {
    SC_HANDLE   schService;
    DWORD       err;

    // Create a new a service object.
    schService = CreateService(SchSCManager,           // handle of service control manager database
                               DriverName,             // address of name of service to start
                               DriverName,             // address of display name
                               SERVICE_ALL_ACCESS,     // type of access to service
                               SERVICE_KERNEL_DRIVER,  // type of service
                               SERVICE_DEMAND_START,   // when to start service
                               SERVICE_ERROR_NORMAL,   // severity if service fails to start
                               ServiceExe,             // address of name of binary file
                               NULL,                   // service does not belong to a group
                               NULL,                   // no tag requested
                               NULL,                   // no dependency names
                               NULL,                   // use LocalSystem account
                               NULL                    // no password for service account
                               );
    if (schService == NULL) {
        err = GetLastError();
        if (err == ERROR_SERVICE_EXISTS) {
            // Ignore this error.
            return TRUE;
        } else {
            printf("CreateService failed!  Error = %d \n", err );
            // Indicate an error.
            return  FALSE;
        }
    }

    // Close the service object.
    if (schService) {
        CloseServiceHandle(schService);
    }

    printf("=== InstallDriver.\n");
    // Indicate success.
    return TRUE;
}   // InstallDriver

BOOLEAN ManageDriver(LPCTSTR DriverName, LPCTSTR ServiceName, USHORT Function) {
    SC_HANDLE   schSCManager;
    BOOLEAN rCode = TRUE;

    // Insure (somewhat) that the driver and service names are valid.
    if (!DriverName || !ServiceName) {
        printf("Invalid Driver or Service provided to ManageDriver() \n");
        return FALSE;
    }

    // Connect to the Service Control Manager and open the Services database.
    schSCManager = OpenSCManager(NULL,                   // local machine
                                 NULL,                   // local database
                                 SC_MANAGER_ALL_ACCESS   // access required
                                 );
    if (!schSCManager) {
        printf("Open SC Manager failed! Error = %d \n", GetLastError());
        return FALSE;
    }

    // Do the requested function.
    switch( Function ) {
        case DRIVER_FUNC_INSTALL:
            // Install the driver service.
            if (InstallDriver(schSCManager, DriverName, ServiceName)) {
                // Start the driver service (i.e. start the driver).
                rCode = StartDriver(schSCManager, DriverName);
            } else {
                // Indicate an error.
                rCode = FALSE;
            }
            break;

        case DRIVER_FUNC_REMOVE:
            // Stop the driver.
            StopDriver(schSCManager, DriverName);
            // Remove the driver service.
            RemoveDriver(schSCManager, DriverName);
            // Ignore all errors.
            rCode = TRUE;
            break;

        default:
            printf("Unknown ManageDriver() function. \n");
            rCode = FALSE;
            break;
    }
    // Close handle to service control manager.

    if (schSCManager) {
        CloseServiceHandle(schSCManager);
    }

    return rCode;
}   // ManageDriver

BOOLEAN RemoveDriver(SC_HANDLE SchSCManager, LPCTSTR DriverName) {
    SC_HANDLE   schService;
    BOOLEAN     rCode;

    // Open the handle to the existing service.
    schService = OpenService(SchSCManager, DriverName, SERVICE_ALL_ACCESS);
    if (schService == NULL) {
        printf("OpenService failed!  Error = %d \n", GetLastError());
        // Indicate error.
        return FALSE;
    }

    // Mark the service for deletion from the service control manager database.
    if (DeleteService(schService)) {
        // Indicate success.
        rCode = TRUE;
    } else {
        printf("DeleteService failed!  Error = %d \n", GetLastError());
        // Indicate failure.  Fall through to properly close the service handle.
        rCode = FALSE;
    }

    // Close the service object.
    if (schService) {
        CloseServiceHandle(schService);
    }

    printf("=== RemoveDriver.\n");
    return rCode;
}   // RemoveDriver

BOOLEAN StartDriver(SC_HANDLE SchSCManager, LPCTSTR DriverName) {
    SC_HANDLE   schService;
    DWORD       err;

    // Open the handle to the existing service.
    schService = OpenService(SchSCManager, DriverName, SERVICE_ALL_ACCESS);
    if (schService == NULL) {
        printf("OpenService failed!  Error = %d \n", GetLastError());
        // Indicate failure.
        return FALSE;
    }

    // Start the execution of the service (i.e. start the driver).
    if (!StartService(schService,     // service identifier
                      0,              // number of arguments
                      NULL            // pointer to arguments
                      )) {
        err = GetLastError();
        if (err == ERROR_SERVICE_ALREADY_RUNNING) {
            // Ignore this error.
            return TRUE;
        } else {
            printf("StartService failure! Error = %d \n", err );
            // Indicate failure.  Fall through to properly close the service handle.
            return FALSE;
        }

    }

    // Close the service object.
    if (schService) {
        CloseServiceHandle(schService);
    }

    printf("=== StartDriver.\n");
    return TRUE;
}   // StartDriver



BOOLEAN StopDriver(SC_HANDLE SchSCManager, LPCTSTR DriverName) {
    BOOLEAN         rCode = TRUE;
    SC_HANDLE       schService;
    SERVICE_STATUS  serviceStatus;

    // Open the handle to the existing service.
    schService = OpenService(SchSCManager, DriverName, SERVICE_ALL_ACCESS);
    if (schService == NULL) {
        printf("OpenService failed!  Error = %d \n", GetLastError());
        return FALSE;
    }

    // Request that the service stop.
    if (ControlService(schService, SERVICE_CONTROL_STOP, &serviceStatus)) {
        // Indicate success.
        rCode = TRUE;
    } else {
        printf("ControlService failed!  Error = %d \n", GetLastError() );
        // Indicate failure.  Fall through to properly close the service handle.
        rCode = FALSE;
    }

    // Close the service object.
    if (schService) {
        CloseServiceHandle (schService);
    }

    return rCode;
}   //  StopDriver

BOOLEAN SetupDriverName(PUCHAR DriverLocation) {
    HANDLE fileHandle;
    DWORD driverLocLen = 0;

    // Get the current directory.
    driverLocLen = GetCurrentDirectory(MAX_PATH, DriverLocation);

    if (!driverLocLen) {
        printf("GetCurrentDirectory failed!  Error = %d \n", GetLastError());
        return FALSE;
    }

    // Setup path name to driver file.
    strcat(DriverLocation, "\\");
    strcat(DriverLocation, DRIVER_NAME);
    strcat(DriverLocation, ".sys");

    // Insure driver file is in the specified directory.
    if ((fileHandle = CreateFile(DriverLocation,
                                 GENERIC_READ,
                                 0,
                                 NULL,
                                 OPEN_EXISTING,
                                 FILE_ATTRIBUTE_NORMAL,
                                 NULL
                                 )) == INVALID_HANDLE_VALUE) {
        printf("Driver: MESAPCI.SYS is not in the current or system directory. \n");
        return FALSE;
    }

    // Close open file handle.
    if (fileHandle) {
        CloseHandle(fileHandle);
    }

    return TRUE;
}   // SetupDriverName

#endif
