/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <sys/stat.h>

#ifndef __WIN32__
  #include "upci.h"
#else
  #include "drvmgr.h"
#endif

#include "mesa5i25.h"

__u8 BootBlock[BOOT_BLOCK_SIZE] = {
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
    0xFF, 0xFF, 0xAA, 0x99, 0x55, 0x66, 0x31, 0xE1,
    0xFF, 0xFF, 0x32, 0x61, 0x00, 0x00, 0x32, 0x81,
    0x0B, 0x08, 0x32, 0xA1, 0x00, 0x00, 0x32, 0xC1,
    0x0B, 0x01, 0x30, 0xA1, 0x00, 0x00, 0x33, 0x01,
    0x21, 0x00, 0x32, 0x01, 0x00, 0x1F, 0x30, 0xA1,
    0x00, 0x0E, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
    0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00};

static int region = -1;
static int dev = -1;
static char PageBuff[SECTOR_SIZE] = {0, };

static inline void wait_for_spi_data(void) {
    int i = 0;
    __u8 data = 0;

//    upci_read_u32(region, SPI_CTRL_REG);
    while ((data & DAV_MASK) == 0) {
        data = upci_read_u32(region, SPI_CTRL_REG);
//        printf("SPI_CTRL_REG 0x%X, DAV_MASK=0x%X I=%d\n", data, DAV_MASK, i);
        if (++i > 5000) {
            printf("timeout waiting for SPI data\n");
            exit(-1);
        }
    }
}

static inline void send_spi_byte(__u32 data) {
    upci_write_u32(region, SPI_DATA_REG, data);
    wait_for_spi_data();
}

static inline __u32 recv_spi_byte(__u32 byte) {
    upci_write_u32(region, SPI_DATA_REG, byte);
    wait_for_spi_data();
    return upci_read_u32(region, SPI_DATA_REG);
}

static inline void put_address(__u32 addr) {
    send_spi_byte((addr >> 16) & 0xFF);
    send_spi_byte((addr >> 8) & 0xFF);
    send_spi_byte(addr & 0xFF);
}

static inline __u32 read_byte25(__u32 addr) {
    __u32 ret;

    PREFIX;
    send_spi_byte(CMD_READ);
    put_address(addr);
    ret = recv_spi_byte(0);
    SUFFIX;
    return ret;
}

static inline __u32 read_status25(void) {
    __u32 ret;

    PREFIX;
    send_spi_byte(CMD_READ_STATUS);
    ret = recv_spi_byte(0);
    SUFFIX;
    return ret;
}

static inline void write_enable(void) {
  PREFIX;
  send_spi_byte(CMD_WRITE_ENABLE);
  SUFFIX;
}

static inline void wait_for_write(void) {
    __u8 stat = read_status25();
  
    while ((stat & WRITE_IN_PROGRESS_MASK) != 0) {
         stat = read_status25();
    }
}

static inline __u32 read_cookie(void) {
    return upci_read_u32(region, COOKIE_REG);
}

static inline __u8 read_ID25(void) {
    __u8 ret;
    PREFIX;
    send_spi_byte(CMD_READ_IDROM);
    put_address(0); // three dummy bytes
    ret = recv_spi_byte(0);
  //  printf("READID25: returned 0x%X\n", ret);
    SUFFIX;
    return ret;
}

/////////////////////////////////////////////////////////

void erase_sector(__u32 addr) {
    write_enable();
    PREFIX;
    send_spi_byte(CMD_SECTOR_ERASE);
    put_address(addr);
    SUFFIX;
    wait_for_write();
}

void write_block(__u32 addr, int buff_i) {
    int i; 

    write_enable();
    PREFIX;
    send_spi_byte(CMD_PAGE_WRITE);
    put_address(addr);// { note that add 0..7 should be 0}
    for (i = 0; i < PAGE_SIZE; i++) {
        send_spi_byte(PageBuff[buff_i + i]);
    }
    SUFFIX;
    wait_for_write();
}

int start_programming(__u32 start_addr, int fsize) {
    __u32 sec_addr;
    int esectors, sector;
    
//  IdentifyEEPROM;
    SET_CS_HIGH;

    esectors = (fsize - 1) / SECTOR_SIZE;
    if (esectors > 7) {
        printf("File Size too large to fit\n");
        return -1;
    }
    sec_addr = start_addr;
    printf("Erasing EEPROM sectors starting from 0x%X...\n", start_addr);
    printf("  |");
    fflush(stdout);
    for (sector = 0; sector <= esectors; sector++) {
        erase_sector(sec_addr);
        sec_addr = sec_addr + SECTOR_SIZE;
        printf("E");
        fflush(stdout);
    }
    printf("\n");
    SET_CS_HIGH;
    return 0;
}

void done_programming(void) {
  SET_CS_HIGH;
}

/////////////////////////////////////////////////////////

int m5i25_print_bitfile_header(FILE *fp) {
    unsigned char buff[256];
    int sleng;
    int bytesread, conflength;

    bytesread = fread(&buff, 1, 14, fp);
    if (bytesread != 14) {
        if (feof(fp)) 
            printf("Unexpected end of file\n");
        else
            printf("IO error\n");
        return -1;
    }
    if ((buff[0] == 0) && (buff[1] == 9)) {
        printf("OK\n");
        printf("  File type: BIT file\n");
        if ((buff[11] == 0) && (buff[12] == 1) && (buff[13] == 0x61)) {
            bytesread = fread(&buff, 1, 2, fp);
            sleng = buff[0]*256 + buff[1];
            bytesread = fread(&buff, 1, sleng, fp);
            printf("  Design name: %s\n", buff);

            bytesread = fread(&buff, 1, 3, fp);
            sleng = buff[1]*256 + buff[2];
            bytesread = fread(&buff, 1, sleng, fp);
            printf("  Part name: %s\n", buff);

            bytesread = fread(&buff, 1, 3, fp);
            sleng = buff[1]*256 + buff[2];
            bytesread = fread(&buff, 1, sleng, fp);
            printf("  Design date: %s\n", buff);

            bytesread = fread(&buff, 1, 3, fp);
            sleng = buff[1]*256 + buff[2];
            bytesread = fread(&buff, 1, sleng, fp);
            printf("  Design time: %s\n", buff);

            bytesread = fread(&buff, 1, 5, fp);
            conflength = buff[1]*16777216;
            conflength = conflength + buff[2]*65536;
            conflength = conflength + buff[3]*256;
            conflength = conflength + buff[4];
            printf("  Config Length: %d\n", conflength);
        }
        return 0;
    }
    if ((buff[0] == 0xFF) && (buff[1] == 0xFF) && (buff[2] == 0xFF) && (buff[3] == 0xFF)) {
        printf("Looks like a BIN file\n");
        fseek(fp, 1, 0);
        return -1;
    }
    printf("Invalid bitfile header!\n");
    return -1;
}

void m5i25_write_area(char *fname, __u32 start_address) {
    int bytesread, i;
    __u32 eeprom_addr;
    struct stat file_stat;
    FILE *fp;
  
    if (m5i25_check_boot() == -1)
        m5i25_write_boot();
    else 
        printf("BootSector OK\n");
    if (m5i25_check_boot() == -1) {
        printf("Failed to write valid boot sector\n");
        exit(-1);
    }

    if (stat(fname, &file_stat) != 0) {
        printf("Can't find file %s\n", fname);
        exit(-1);
    }

    printf("Checking file %s... ", fname);
    fp = fopen(fname, "rb");
        
    if (m5i25_print_bitfile_header(fp) == -1) {
        fclose(fp);
        exit(-1);
    }
    
    if (start_programming(start_address, file_stat.st_size) == -1) {
        fclose(fp);
        exit(-1);
    }
    printf("Programming EEPROM area starting from 0x%X...\n", start_address);
    printf("  |");
    fflush(stdout);
    eeprom_addr = start_address;
    while (!feof(fp)) {
      bytesread = fread(&PageBuff, 1, 8192, fp);
      i = 0;
      printf("W");
      fflush(stdout);
      while (i < bytesread) {
        write_block(eeprom_addr, i);
        i = i + PAGE_SIZE;
        eeprom_addr = eeprom_addr + PAGE_SIZE;
      }
    }
    printf("\nBoard configuration updated successfully\n");
    done_programming();
}

void m5i25_verify_area(char *fname, __u32 start_address) {
    int bytesread, i, bindex;
    __u8 rdata, mdata;
    struct stat file_stat;
    FILE *fp;
  
    if (m5i25_check_boot() == -1) {
        printf("ERROR: BootSector is invalid\n");
    }

    if (stat(fname, &file_stat) != 0) {
        printf("Can't find file %s\n", fname);
        exit(-1);
    }

    printf("Checking file %s... ", fname);
    fp = fopen(fname, "rb");
        
    if (m5i25_print_bitfile_header(fp) == -1) {
        fclose(fp);
        exit(-1);
    }

    printf("  |");
    i = start_address;
    while (!feof(fp)) {
      bytesread = fread(&PageBuff, 1, 8192, fp);
      bindex = 0;
      while (bindex < bytesread) {
        rdata = read_byte25(bindex + i);
        mdata = PageBuff[bindex];
        if (mdata != rdata) {
          printf("\nError at 0x%X expected: 0x%X but read: 0x%X\n", i+bindex, mdata, rdata);
          exit(0);
        }
        bindex = bindex + 1;
    }
    i = i + 8192;
    printf("V");
    fflush(stdout);
    }

    printf("\nBoard configuration verified successfully\n");
}

int m5i25_check_cookie(void) {
    __u32 cookie;

    printf("Checking cookie... ");
    cookie = read_cookie();
    if (cookie != COOKIE_5i25) {
        printf("ERROR: Invalid cookie: %x, expected %x\n", cookie, COOKIE_5i25);
        return -1;
    } else {
        printf("OK\n");
        return 0;
    }
}    

int m5i25_identify_EEPROM(void) {
    __u8 id;
    
    printf("Identifing EEPROM... ");
    id = read_ID25();

    switch (id) {
        case ID_EEPROM_1M:
            printf("OK\n  1Mb found\n"); 
            break;
        case ID_EEPROM_2M:
            printf("OK\n  2Mb found\n"); 
            break;
        case ID_EEPROM_4M:
            printf("OK\n  4Mb found\n"); 
            break;
        case ID_EEPROM_8M:
            printf("OK\n  8Mb found\n"); 
            break;

        default:
            printf("ERROR: wrong EEPROM id: %x\n", id); 
            return -1;
    } 
    return 0;
}

void m5i25_write_boot(void) {
    printf("Erasing sector 0 for boot block\n");
    erase_sector(BOOT_ADDRESS);
    memset(PageBuff, 0, PAGE_SIZE);
    memcpy(&PageBuff, &BootBlock, BOOT_BLOCK_SIZE);
    write_block(0, 0);
    printf("BootBlock installed\n");
}

int m5i25_check_boot(void) {
    int i, ret = 0;
    __u8 data;

    for (i = 0; i < BOOT_BLOCK_SIZE; i++) {
        data = read_byte25(i);
//        printf("0x%X - %X\n", data, BootBlock[i]);
        if (data != BootBlock[i]) {
            ret = -1;
            break;
        }
    }
    return ret;
}

int m5i25_open_region(int _reg) {
#ifndef __WIN32__
    region = upci_open_region(dev, 0);
    return region;
#else
    HANDLE hDriver = getHandle();
    DWORD dwBytes;
    DWORD ret;
          
    ret = DeviceIoControl(hDriver, IOCTL_MESA_MAP, NULL, 0, NULL, 0, &dwBytes, NULL);
    printf("=== Memory mapped\n");

    return 0;

#endif
}

void m5i25_close_region(void) {
#ifndef __WIN32__
    upci_close_region(region);
#else
    HANDLE hDriver = getHandle();
    DWORD dwBytes;
    DWORD ret;
          
    ret = DeviceIoControl(hDriver, IOCTL_MESA_UNMAP, NULL, 0, NULL, 0, &dwBytes, NULL);
    printf("=== Memory unmapped\n");

#endif
}

int detect_board(__u16 vendor_id, __u16 device_id) {
#ifndef __WIN32__

    struct upci_dev_info info;

    info.vendor_id = vendor_id;
    info.device_id = device_id;
    info.instance = 0;
    dev = upci_find_device(&info);
    return dev;
#else
    HANDLE hDriver = getHandle();
    DWORD dwBytes, detected;
    detect_board_msg db;
    db.vendor = vendor_id;
    db.device = device_id;
    DWORD ret;
          
    ret = DeviceIoControl(hDriver, IOCTL_MESA_DETECT, &db, sizeof(db), &detected, sizeof(DWORD), &dwBytes, NULL);

    if (detected == 0) {
        return -1;
    } else {
        return 0;
    }
#endif
}
