/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#ifdef __WIN32__
#else
  #include <termios.h>
  #include "upci.h"
#endif

#include "mesa5i25.h"

void usage(void) {
   printf("Syntax: mesaflash filename command\n");
   printf("Writes config data to 5I25/6I25 EEPROM\n");
   printf("  filename - VALID FPGA configuration file.\n");
   printf("Commands:\n");
   printf("  write - write program in 'filename' into FPGA.\n");
   printf("  verify - verify if 'filename' is properly programmed in FPGA.\n");
   printf("  info - print info about 'filename'.\n");
   printf("  detect - print info about detected board(s) ('filename' is ignored).\n");
}

int main(int argc, char *argv[])
{
    char buff[255];
    char *fname = &buff[0];
    int mode = 0;
    int board5i25, board6i25, bridge6i25;
        
    if (argc == 3) {
        strncpy(fname, argv[1], 254);
        if (strncmp("write\0", argv[2], 6) == 0) {
            mode = 1;
        } else if (strncmp("verify\0", argv[2], 7) == 0) {
            mode = 2;
        } else  if (strncmp("fallback\0", argv[2], 9) == 0) {
            mode = 3;
        } else  if (strncmp("info\0", argv[2], 5) == 0) {
            mode = 4;
        } else  if (strncmp("detect\0", argv[2], 7) == 0) {
            mode = 5;
        } else {
            printf("Unknown command line option: %s\n", argv[2]);
            return 0;
        }
    } else {
        usage();
        return 0;
    }

    if (mode == 4) {
        struct stat file_stat;
        FILE *fp;
  
        if (stat(fname, &file_stat) != 0) {
            printf("Can't find file %s\n", fname);
            exit(-1);
        }

        printf("Checking file %s...", fname);
        fp = fopen(fname, "r");
        
        m5i25_print_bitfile_header(fp);
        fclose(fp);
        return 0;
    }

    upci_reset();
    if (upci_scan_bus() == -1)
        return -1;

    bridge6i25 = detect_board(HM2_PCI_VENDORID_XIO2001, HM2_PCI_DEVICEID_XIO2001);
    board6i25 = detect_board(HM2_PCI_VENDORID_MESAPCI, HM2_PCI_DEVICEID_MESA6I25);
    board5i25 = detect_board(HM2_PCI_VENDORID_MESAPCI, HM2_PCI_DEVICEID_MESA5I25);

    if ((bridge6i25 != -1) && (board6i25 != -1)) {
        printf("Found mesa6i25 board\n");
        if (mode == 5) {
            upci_print_device_info(bridge6i25);
            upci_print_device_info(board6i25);
        }
    } else if ((bridge6i25 != -1) && (board5i25 != -1)) {
        printf("Found mesa6i25 board with 5i25 flash\n");
        if (mode == 5) {
            upci_print_device_info(bridge6i25);
            upci_print_device_info(board5i25);
        }
    } else if ((bridge6i25 == -1) && (board6i25 != -1)) {
        printf("Found mesa5i25 board with 6i25 flash\n");
        if (mode == 5) {
            upci_print_device_info(board6i25);
        }
    } else if ((bridge6i25 == -1) && (board5i25 != -1)) {
        printf("Found mesa5i25 board\n");
        if (mode == 5) {
            upci_print_device_info(board5i25);
        }
    } else {
        printf("Board not found\n");
        return -1;
    }

    if (mode == 5)
        return 0;

    if (m5i25_open_region(0) == -1)
        return -1;

    atexit(m5i25_close_region);

    if (m5i25_check_cookie() == -1)
        return -1;

    if (m5i25_identify_EEPROM() == -1)
        return -1;

    if (mode == 1) {
        m5i25_write_area(fname, USER_ADDRESS);
    } else if (mode == 2) {
        m5i25_verify_area(fname, USER_ADDRESS);
    } else if (mode == 3) {
        m5i25_write_area(fname, FALLBACK_ADDRESS);
    }

    return 0;
}
