
#ifdef __linux__
    #include <pci/pci.h>
    #include <sys/socket.h>
    #include <netinet/in.h>
    #include <arpa/inet.h>
    #include <unistd.h>
    #include <time.h>
#elif _WIN32
    #include "libpci/pci.h"
    #include <windows.h>
    #include <winsock2.h>
#endif
#include <fcntl.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <math.h>

#include "eeprom.h"
#include "hostmot2.h"
#include "lbp16.h"
#include "device_eth.h"

#define PORT 27181

#ifdef __linux__
    int sd;
    socklen_t len;
#elif _WIN32
    SOCKET sd;
    int len;
#endif
struct sockaddr_in server_addr, client_addr;

static u8 page_from_eth[PAGE_SIZE] = {0, };
extern u8 file_buffer[SECTOR_SIZE];
extern u8 boot_block[BOOT_BLOCK_SIZE];

static u8 read_idrom_eth(void) {
    return (u8) lbp16_send_read_u32(CMD_READ_FLASH_IDROM, FLASH_ID_REG);
}

static u32 read_cookie_eth(void) {
    return lbp16_send_read_u32(CMD_READ_HM2_COOKIE, HM2_COOKIE_REG);
}

static void read_page(u32 addr, void *buff) {
    lbp16_send_flash_address(addr);
    lbp16_send_flash_page_read(buff);
}

static void write_page(u32 addr, void *buff) {
    lbp16_send_flash_address(addr);
    lbp16_send_flash_page_write(buff);
}

static void write_boot_eth(void) {
    printf("Erasing sector 0 for boot block\n");
    lbp16_send_sector_erase(BOOT_ADDRESS);
    memset(&file_buffer, 0, PAGE_SIZE);
    memcpy(&file_buffer, &boot_block, BOOT_BLOCK_SIZE);
    lbp16_send_flash_page_write(file_buffer);
    printf("BootBlock installed\n");
}

static int check_boot_eth(void) {
    int i;

    read_page(0x0, &page_from_eth);
    for (i = 0; i < BOOT_BLOCK_SIZE; i++) {
        if (boot_block[i] != page_from_eth[i]) {
            return -1;
        }
    }
    return 0;
}

static int start_programming(mesa_board_t *board, u32 start_address, int fsize) {
    u32 sec_addr;
    int esectors, sector, max_sectors;

    esectors = (fsize - 1) / SECTOR_SIZE;
    if (start_address == FALLBACK_ADDRESS) {
        max_sectors = calc_user_space(board->flash_id) / SECTOR_SIZE - 1;
    } else {
        max_sectors = calc_user_space(board->flash_id) / SECTOR_SIZE;
    }
    if (esectors > max_sectors) {
        printf("File Size too large to fit\n");
        return -1;
    }
    printf("EEPROM sectors to write: %d, max sectors in area: %d\n", esectors, max_sectors);
    sec_addr = start_address;
    printf("Erasing EEPROM sectors starting from 0x%X...\n", (unsigned int) start_address);
    printf("  |");
    fflush(stdout);
    for (sector = 0; sector <= esectors; sector++) {
        lbp16_send_sector_erase(sec_addr);
        sec_addr = sec_addr + SECTOR_SIZE;
        printf("E");
        fflush(stdout);
    }
    printf("\n");
    return 0;
}

// public functions

void eth_print_all_info(mesa_board_t *board) {
    lbp16_cmd_addr packet;
    u8 buff[20];
    int i, j;
    u8 idrom;
    char *mem_types[16] = {NULL, "registers", "memory", NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, "EEPROM", "flash"};
    char *mem_writeable[2] = {"RO", "RW"};
    char *acc_types[4] = {"8bit", "16bit", "32bit", "64bit"};
    char *led_debug_types[2] = {"Hostmot2", "eth debug"};
    char *boot_jumpers_types[4] = {"fixed 192.168.1.121", "fixed from EEPROM", "from BOOTP", "INVALID"};
    lbp_mem_info_area mem_area;
    lbp_eth_eeprom_area eth_area;
    lbp_status_area stat_area;
    lbp_info_area info_area;
    lbp16_cmd_addr cmds[LBP16_MEM_SPACE_COUNT];

    LBP16_INIT_PACKET4(packet, CMD_READ_ETH_EEPROM_ADDR16_INCR(sizeof(eth_area)/2), 0);
    memset(&eth_area, 0, sizeof(eth_area));
    sendto (sd, (char*) &packet, sizeof(packet), 0, (struct sockaddr *) &server_addr, sizeof(server_addr));
    recvfrom (sd, (char*) &eth_area, sizeof (eth_area), 0, (struct sockaddr *) &client_addr, &len);

    LBP16_INIT_PACKET4(packet, CMD_READ_COMM_CTRL_ADDR16_INCR(sizeof(stat_area)/2), 0);
    memset(&stat_area, 0, sizeof(stat_area));
    sendto (sd, (char*) &packet, sizeof(packet), 0, (struct sockaddr *) &server_addr, sizeof(server_addr));
    recvfrom (sd, (char*) &stat_area, sizeof (stat_area), 0, (struct sockaddr *) &client_addr, &len);

    LBP16_INIT_PACKET4(packet, CMD_READ_BOARD_INFO_ADDR16_INCR(sizeof(info_area)/2), 0);
    memset(&info_area, 0, sizeof(info_area));
    sendto (sd, (char*) &packet, sizeof(packet), 0, (struct sockaddr *) &server_addr, sizeof(server_addr));
    recvfrom (sd, (char*) &info_area, sizeof (info_area), 0, (struct sockaddr *) &client_addr, &len);

    if (board->verbose != 0) {
        LBP16_INIT_PACKET4(packet, CMD_READ_BOARD_INFO_ADDR16_INCR(8), 0);
        memset(buff, 0, sizeof(buff));
        sendto (sd, (char*) &packet, sizeof(packet), 0, (struct sockaddr *) &server_addr, sizeof(server_addr));
        recvfrom (sd, (char*) &buff, sizeof (buff), 0, (struct sockaddr *) &client_addr, &len);
        printf("board name: %s\n", buff);

        LBP16_INIT_PACKET4(cmds[0], CMD_READ_AREA_INFO_ADDR16_INCR(LBP_SPACE_HM2, sizeof(mem_area)/2), 0);
        LBP16_INIT_PACKET4(cmds[1], CMD_READ_AREA_INFO_ADDR16_INCR(LBP_SPACE_ETH_CHIP, sizeof(mem_area)/2), 0);
        LBP16_INIT_PACKET4(cmds[2], CMD_READ_AREA_INFO_ADDR16_INCR(LBP_SPACE_ETH_EEPROM, sizeof(mem_area)/2), 0);
        LBP16_INIT_PACKET4(cmds[3], CMD_READ_AREA_INFO_ADDR16_INCR(LBP_SPACE_FPGA_FLASH, sizeof(mem_area)/2), 0);
        LBP16_INIT_PACKET4(cmds[4], 0, 0);
        LBP16_INIT_PACKET4(cmds[5], 0, 0);
        LBP16_INIT_PACKET4(cmds[6], CMD_READ_AREA_INFO_ADDR16_INCR(LBP_SPACE_COMM_CTRL, sizeof(mem_area)/2), 0);
        LBP16_INIT_PACKET4(cmds[7], CMD_READ_AREA_INFO_ADDR16_INCR(LBP_SPACE_BOARD_INFO, sizeof(mem_area)/2), 0);

        for (i = 0; i < 8; i++) {
            if ((cmds[i].cmd_lo == 0) && (cmds[i].cmd_hi == 0)) continue;
            memset(&mem_area, 0, sizeof(mem_area));
            sendto (sd, (char*) &cmds[i], sizeof(cmds[i]), 0, (struct sockaddr *) &server_addr, sizeof(server_addr));
            recvfrom (sd, (char*) &mem_area, sizeof (mem_area), 0, (struct sockaddr *) &client_addr, &len);

            printf("area %d: %.*s\n", i, sizeof(mem_area.name), mem_area.name);
            printf("  %s, %s,", mem_writeable[(mem_area.size & 0x8000) >> 15], mem_types[(mem_area.size  >> 8) & 0x7F]);
            for (j = 0; j < 4; j++) {
                if ((mem_area.size & 0xFF) & 1 << j)
                    printf(" %s", acc_types[j]);
            }
            printf(" [size=%u]", (unsigned int) pow(2, mem_area.range & 0x3F));
            if (((mem_area.size  >> 8) & 0x7F) >= 0xE)
                printf(", page size: 0x%X, erase size: 0x%X\n",
                  (unsigned int) pow(2, (mem_area.range >> 6) & 0x1F), (unsigned int) pow(2, (mem_area.range >> 11) & 0x1F));
            else
                printf("\n");
        }
    }

    if (board->verbose != 0) {
        printf("eth eeprom info:\n");
        printf("  mac address: %02X:%02X:%02X:%02X:%02X:%02X\n", HI_BYTE(eth_area.mac_addr_hi), LO_BYTE(eth_area.mac_addr_hi),
          HI_BYTE(eth_area.mac_addr_mid), LO_BYTE(eth_area.mac_addr_mid), HI_BYTE(eth_area.mac_addr_lo), LO_BYTE(eth_area.mac_addr_lo));
        if (info_area.jumpers == 1) {
            printf("  ip address: %d.%d.%d.%d\n", HI_BYTE(eth_area.ip_addr_hi), LO_BYTE(eth_area.ip_addr_hi),
              HI_BYTE(eth_area.ip_addr_lo), LO_BYTE(eth_area.ip_addr_lo));
        } else if (info_area.jumpers == 0) {
            printf("  ip address: 192.168.1.121\n");
        }
        printf("  board name: %.*s\n", sizeof(eth_area.name), eth_area.name);
    } else {
        printf("board name: %.*s\n", sizeof(eth_area.name), eth_area.name);
        if (info_area.jumpers == 1) {
            printf("  ip address: %d.%d.%d.%d\n", HI_BYTE(eth_area.ip_addr_hi), LO_BYTE(eth_area.ip_addr_hi),
              HI_BYTE(eth_area.ip_addr_lo), LO_BYTE(eth_area.ip_addr_lo));
        } else if (info_area.jumpers == 0) {
            printf("  ip address: 192.168.1.121\n");
        }
    }

    if (board->verbose != 0) {
        printf("  user leds: %s\n", led_debug_types[eth_area.led_debug & 0x1]);

        printf("FPGA flash eeprom info:\n");
        idrom = read_idrom();
        printf("  flash id: 0x%02X %s\n", idrom, print_idrom_type(idrom));

        printf("lba16 status/control info:\n");
        printf("  packets recived: all %d, UDP %d, bad %d\n", stat_area.RXPacketCount, stat_area.RXUDPCount, stat_area.RXBadCount);
        printf("  packets sended: all %d, UDP %d, bad %d\n", stat_area.TXPacketCount, stat_area.TXUDPCount, stat_area.TXBadCount);
        printf("  parse errors: %d, mem errors %d, write errors %d\n", stat_area.LBPParseErrors, stat_area.LBPMemErrors, stat_area.LBPWriteErrors);
        printf("  UDP time: %d\n", stat_area.UDPPktTimeStamp);

        printf("board info:\n");
        printf("  board name: %.*s\n", sizeof(info_area.name), info_area.name);
        printf("  LBP16 version %d\n", info_area.LBP_version);
        printf("  firmware version %d\n", info_area.firmware_version);
        printf("  IP address jumpers at boot: %s\n", boot_jumpers_types[info_area.jumpers]);
    }
}

void eth_print_firmware_info(void) {
    lbp16_cmd_addr packet;
    u8 name[10];
    u32 idrom_addr, mod_addr, idrom_ver;
    u32 mod_desc;
    int i;

    memset(name, 0, sizeof(name));
    printf("firmware info:\n");
    printf("  cookie: %X\n", (unsigned int) read_cookie());
    LBP16_INIT_PACKET4(packet, CMD_READ_HOSTMOT2_ADDR32_INCR(2), HM2_COOKIE_REG + 4);
    sendto (sd, (char*) &packet, sizeof(packet), 0, (struct sockaddr *) &server_addr, sizeof(server_addr));
    recvfrom (sd, (char*) name, sizeof(name), 0, (struct sockaddr *) &client_addr, &len);
    printf("  name: %s\n", name);
    LBP16_INIT_PACKET4(packet, CMD_READ_HOSTMOT2_ADDR32(1), HM2_COOKIE_REG + 12);
    sendto (sd, (char*) &packet, sizeof(packet), 0, (struct sockaddr *) &server_addr, sizeof(server_addr));
    recvfrom (sd, (char*) &idrom_addr, sizeof(idrom_addr), 0, (struct sockaddr *) &client_addr, &len);

    LBP16_INIT_PACKET4(packet, CMD_READ_HOSTMOT2_ADDR32(1), idrom_addr);
    sendto (sd, (char*) &packet, sizeof(packet), 0, (struct sockaddr *) &server_addr, sizeof(server_addr));
    recvfrom (sd, (char*) &idrom_ver, sizeof(idrom_ver), 0, (struct sockaddr *) &client_addr, &len);
    printf("  IDROM version: %u\n", (unsigned int) idrom_ver);

    LBP16_INIT_PACKET4(packet, CMD_READ_HOSTMOT2_ADDR32(1), idrom_addr + 4);
    sendto (sd, (char*) &packet, sizeof(packet), 0, (struct sockaddr *) &server_addr, sizeof(server_addr));
    recvfrom (sd, (char*) &mod_addr, sizeof(mod_addr), 0, (struct sockaddr *) &client_addr, &len);
    mod_addr += idrom_addr;

    for (i = 0; i < 31; i++) {
        u8 gtag, version, instances;

        LBP16_INIT_PACKET4(packet, CMD_READ_HOSTMOT2_ADDR32(1), mod_addr);
        sendto (sd, (char*) &packet, sizeof(packet), 0, (struct sockaddr *) &server_addr, sizeof(server_addr));
        recvfrom (sd, (char*) &mod_desc, sizeof(mod_desc), 0, (struct sockaddr *) &client_addr, &len);

        gtag = mod_desc & 0xFF;
        version = (mod_desc >> 8) & 0xFF;
        instances = (mod_desc >> 24) & 0xFF;
        if (gtag == 0) break;
        printf("  Module %d: %s, version %d [instances=%d]\n", i, hm2_get_general_function_name(gtag), version, instances);
        mod_addr += 12;
    }
}

int verify_area_eth(char *fname, u32 start_address) {
    int bytesread, fl_addr, bindex;
    char part_name[32];
    struct stat file_stat;
    FILE *fp;
    int i;

    if (stat(fname, &file_stat) != 0) {
        printf("Can't find file %s\n", fname);
        return -1;
    }

    fp = fopen(fname, "rb");
    if (fp == NULL) {
        printf("Can't open file %s: %s\n", fname, strerror(errno));
        return -1;
    }
    if (print_bitfile_header(fp, (char*) &part_name) == -1) {
        fclose(fp);
        return -1;
    }
/*    if (strcmp(part_name, active_board.fpga_part_number) != 0) {
        printf("Error: wrong bitfile destination device: %s, should be %s\n", part_name, active_board.fpga_part_number);
        fclose(fp);
        return -1;
    }*/
    if (check_boot_eth() == -1) {
        printf("Error: BootSector is invalid\n");
        fclose(fp);
        return -1;
    } else {
        printf("Boot sector OK\n");
    }

    printf("Verifying EEPROM sectors starting from 0x%X...\n", (unsigned int) start_address);
    printf("  |");
    fflush(stdout);
    fl_addr = start_address;
    while (!feof(fp)) {
        bytesread = fread(&file_buffer, 1, 8192, fp);
        bindex = 0;
        while (bindex < bytesread) {
            read_page(fl_addr, &page_from_eth);
            for (i = 0; i < PAGE_SIZE; i++, bindex++) {
                if (file_buffer[bindex] != page_from_eth[i]) {
                   printf("\nError at 0x%X expected: 0x%X but read: 0x%X\n", bindex, file_buffer[bindex], page_from_eth[i]);
                   return -1;
                }
            }
            fl_addr += PAGE_SIZE;
        }
        printf("V");
        fflush(stdout);
    }

    fclose(fp);
    printf("\nBoard configuration verified successfully\n");
    return 0;
}

int write_area_eth(mesa_board_t *board, char *fname, u32 start_address) {
    int bytesread, i, bindex;
    char part_name[32];
    struct stat file_stat;
    FILE *fp;

    if (stat(fname, &file_stat) != 0) {
        printf("Can't find file %s\n", fname);
        return -1;
    }

    fp = fopen(fname, "rb");
    if (fp == NULL) {
        printf("Can't open file %s: %s\n", fname, strerror(errno));
        return -1;
    }
    if (print_bitfile_header(fp, (char*) &part_name) == -1) {
        fclose(fp);
        return -1;
    }
/*    if (strcmp(part_name, active_board.fpga_part_number) != 0) {
        printf("Error: wrong bitfile destination device: %s, should be %s\n", part_name, active_board.fpga_part_number);
        fclose(fp);
        return -1;
    }*/
    if (check_boot_eth() == -1) {
        write_boot_eth();
    } else {
        printf("Boot sector OK\n");
    }
    if (check_boot_eth() == -1) {
        printf("Failed to write valid boot sector\n");
        fclose(fp);
        return -1;
    }

    if (start_programming(board, start_address, file_stat.st_size) == -1) {
        fclose(fp);
        return -1;
    }
    printf("Programming EEPROM sectors starting from 0x%X...\n", (unsigned int) start_address);
    printf("  |");
    fflush(stdout);
    i = start_address;
    while (!feof(fp)) {
        bytesread = fread(&file_buffer, 1, 8192, fp);
        bindex = 0;
        while (bindex < bytesread) {
            write_page(i, &file_buffer[bindex]);
            i += PAGE_SIZE;
            bindex += PAGE_SIZE;
        }
        printf("W");
        fflush(stdout);
    }

    fclose(fp);
    printf("\nBoard configuration verified successfully\n");
    return 0;
}

int find_all_mesaeth_devices(char *addr) {
    lbp16_cmd_addr packet, packet2;
    int i;
    int send = 0, recv = 0;
    char *ptr;
    u32 cookie;
    u8 buff[20];
#ifdef __linux__
    int val;
    struct timespec tv, tvret;    // for nanosleep in linux, on windows there is only Sleep(ms)
#elif _WIN32
    int iResult;
    WSADATA wsaData;
    u_long iMode = 1;    // block mode in windows
#endif

// open socket
#ifdef __linux__
    sd = socket (PF_INET, SOCK_DGRAM, 0);
#elif _WIN32
    iResult = WSAStartup(MAKEWORD(2,2), &wsaData);
    if (iResult != 0) {
        printf("WSAStartup failed: %d\n", iResult);
        return -1;
    }

    sd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (sd == 0) {
        printf("Can't open socket: %d %d\n", sd, errno);
        return -1;
    }
#endif
    server_addr.sin_family = AF_INET;
    server_addr.sin_port = htons(PORT);
    len = sizeof(client_addr);

// set socket non blocking and setup sleep time
#ifdef __linux__
    val = fcntl(sd, F_GETFL);
    val = val | O_NONBLOCK;
    fcntl(sd, F_SETFL, val);

    tv.tv_sec = 0;
    tv.tv_nsec = 5000000;
#elif _WIN32
    ioctlsocket(sd, FIONBIO, &iMode);
#endif

    ptr = strrchr(addr, '.');
    *ptr = '\0';
    printf("Scanning addresses %s.0/255.255.255.0\n", addr);

    LBP16_INIT_PACKET4(packet, CMD_READ_HM2_COOKIE, HM2_COOKIE_REG);
    for (i = 1; i < 256; i++) {
        char addr_name[32];

        sprintf(addr_name, "%s.%d", addr, i);
        server_addr.sin_addr.s_addr = inet_addr(addr_name);
        send = sendto(sd, (char*) &packet, sizeof(packet), 0, (struct sockaddr *) &server_addr, sizeof(server_addr));
#ifdef __linux__
        nanosleep(&tv, &tvret);
#elif _WIN32
        Sleep(1);
#endif
        recv = recvfrom(sd, (char*) &cookie, sizeof(cookie), 0, (struct sockaddr *) &client_addr, &len);

        if ((recv != -1) && (cookie == HM2_COOKIE)) {
// set socket blocking
#ifdef __linux__
            val = fcntl(sd, F_GETFL);
            val = val & ~O_NONBLOCK;
            fcntl(sd, F_SETFL, val);
#elif _WIN32
            iMode = 0;
            ioctlsocket(sd, FIONBIO, &iMode);
#endif
            LBP16_INIT_PACKET4(packet2, CMD_READ_BOARD_INFO_ADDR16_INCR(8), 0);
            memset(buff, 0, sizeof(buff));
            sendto (sd, (char*) &packet2, sizeof(packet2), 0, (struct sockaddr *) &client_addr, sizeof(client_addr));
            recvfrom (sd, (char*) &buff, sizeof (buff), 0, (struct sockaddr *) &client_addr, &len);
            printf("board name: %s\n", buff);
            printf("  ip address: %s\n", inet_ntoa(client_addr.sin_addr));
// set socket non-blocking
#ifdef __linux__
            val = fcntl(sd, F_GETFL);
            val = val | O_NONBLOCK;
            fcntl(sd, F_SETFL, val);
#elif _WIN32
            iMode = 1;
            ioctlsocket(sd, FIONBIO, &iMode);
#endif
        }
    }
    close(sd);
    return 0;
}

int open_eth_connection(mesa_board_t *board) {
#ifdef __linux__
    sd = socket (PF_INET, SOCK_DGRAM, 0);
#elif _WIN32
    int iResult;
    WSADATA wsaData;

    iResult = WSAStartup(MAKEWORD(2,2), &wsaData);
    if (iResult != 0) {
        printf("WSAStartup failed: %d\n", iResult);
    return -1;
    }

    sd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (sd == 0) {
        printf("Can't open socket: %d %d\n", sd, errno);
    return -1;
    }
#endif
    server_addr.sin_family = AF_INET;
    server_addr.sin_port = htons(PORT);
    server_addr.sin_addr.s_addr = inet_addr(board->ip_addr);
    len = sizeof(client_addr);

    read_cookie = &read_cookie_eth;
    read_idrom = &read_idrom_eth;
    verify_area = &verify_area_eth;
    write_area = &write_area_eth;

    board->fpga_part_number = "6slx16ftg256";
    board->flash_id = read_idrom();
    board->user_address = calc_user_space(board->flash_id);
    prepare_boot_block(board->flash_id);

    eth_print_all_info(board);
    if (board->verbose != 0)
        eth_print_firmware_info();

    return 0;
}

int close_eth_connection(mesa_board_t *board) {
#ifdef __linux__
            shutdown(sd, SHUT_RDWR);
#elif _WIN32
#endif
    return 0;
}
