
#ifdef __linux__
    #include <pci/pci.h>
    #include <sys/mman.h>
    #include <unistd.h>
#elif _WIN32
    #include "libpci/pci.h"
    #include <windows.h>
    #include "winio/winio_nt.h"
#endif
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#include "device_pci.h"
#include "hostmot2.h"
#include "eeprom.h"
#include "flash_spi.h"

extern u8 file_buffer[SECTOR_SIZE];
extern u8 boot_block[BOOT_BLOCK_SIZE];

#ifdef __linux__
    static int memfd = -1;
#elif _WIN32
    typedef struct tagPhysStruct tagPhysStruct_t;
    typedef PBYTE (*winio32_MapPhysToLin_ptr) (tagPhysStruct_t*);
    typedef BOOL (*winio32_UnmapPhysicalMemory_ptr) (tagPhysStruct_t*);
    HMODULE iodll = NULL;
    tagPhysStruct_t phys;
#endif
void *mapped_ptr = NULL;

static u32 map_size = 0;
static u32 map_base_addr = 0;

static int check_boot_spi(void) {
    int i;
    u8 data;

    for (i = 0; i < BOOT_BLOCK_SIZE; i++) {
        data = read_byte(i);
        if (data != boot_block[i]) {
            return -1;
        }
    }
    return 0;
}

static void pci_print_firmware_info(void) {
    u8 name[10];
    u32 *ptr = (u32 *) &name[0];
    u32 idrom_addr, mod_addr, idrom_ver;
    u32 mod_desc;
    int i;

    memset(name, 0, sizeof(name));
    printf("firmware info:\n");
    printf("  cookie: %X\n", (unsigned int) read_cookie());
    *ptr = PCI_READ_MEM32(HM2_COOKIE_REG + 4);
    ptr = (u32 *) &name[4];
    *ptr = PCI_READ_MEM32(HM2_COOKIE_REG + 8);
    printf("  name: %s\n", name);
    idrom_addr = PCI_READ_MEM32(HM2_COOKIE_REG + 12);
    idrom_ver = PCI_READ_MEM32(idrom_addr);
    printf("  IDROM version: %u\n", (unsigned int) idrom_ver);
    mod_addr = idrom_addr + PCI_READ_MEM32(idrom_addr + 4);
    for (i = 0; i < 31; i++) {
        u8 gtag, version, instances;

        mod_desc = PCI_READ_MEM32(mod_addr);
        gtag = mod_desc & 0xFF;
        version = (mod_desc >> 8) & 0xFF;
        instances = (mod_desc >> 24) & 0xFF;
        if (gtag == 0) break;
        printf("  Module %d: %s, version %d [instances=%d]\n", i, hm2_get_general_function_name(gtag), version, instances);
        mod_addr += 12;
    }
}

static void write_boot_spi(void) {
    printf("Erasing sector 0 for boot block\n");
    erase_sector(BOOT_ADDRESS);
    memset(&file_buffer, 0, PAGE_SIZE);
    memcpy(&file_buffer, &boot_block, BOOT_BLOCK_SIZE);
    write_page(0, (char *) &file_buffer, 0);
    printf("BootBlock installed\n");
}

static int start_programming(mesa_board_t *board, u32 start_address, int fsize) {
    u32 sec_addr;
    int esectors, sector, max_sectors;

    set_cs_high();

    esectors = (fsize - 1) / SECTOR_SIZE;
    if (start_address == FALLBACK_ADDRESS) {
        max_sectors = calc_user_space(board->flash_id) / SECTOR_SIZE - 1;
    } else {
        max_sectors = calc_user_space(board->flash_id) / SECTOR_SIZE;
    }
    if (esectors > max_sectors) {
        printf("File Size too large to fit\n");
        return -1;
    }
    printf("EEPROM sectors to write: %d, max sectors in area: %d\n", esectors, max_sectors);
    sec_addr = start_address;
    printf("Erasing EEPROM sectors starting from 0x%X...\n", (unsigned int) start_address);
    printf("  |");
    fflush(stdout);
    for (sector = 0; sector <= esectors; sector++) {
        erase_sector(sec_addr);
        sec_addr = sec_addr + SECTOR_SIZE;
        printf("E");
        fflush(stdout);
    }
    printf("\n");
    set_cs_high();
    return 0;
}

static u32 read_cookie_fpga(void) {
    return PCI_READ_MEM32(HM2_COOKIE_REG);
}

///////

void pci_print_all_info(mesa_board_t *board, int verbose) {
    u8 idrom;
    struct pci_dev *dev = board->dev;
    char *area_types[2] = {"Memory", "I/O"};

    if (dev->vendor_id == VENDORID_MESAPCI) {
        if (dev->device_id == DEVICEID_MESA5I25)
            printf("board name: 5I25\n");
        else if (dev->device_id == DEVICEID_MESA6I25)
            printf("board name: 6I25\n");
        printf("  PCI device: %02X:%02X.%X (%04X:%04X)\n", dev->bus, dev->dev, dev->func, dev->vendor_id, dev->device_id);
        printf("  Region 0: %s at %08X [size=%u]\n", area_types[dev->base_addr[0] & 0x1], (unsigned int) dev->base_addr[0], (unsigned int) dev->size[0]);
        if ((read_idrom != NULL) && (verbose != 0)) {
            idrom = read_idrom();
            printf("  flash id: 0x%02X %s\n", idrom, print_idrom_type(idrom));
        }
    }
}

int verify_area_pci(char *fname, u32 start_address) {
    int bytesread, i, bindex;
    u8 rdata, mdata;
    char part_name[32];
    struct stat file_stat;
    FILE *fp;

    if (stat(fname, &file_stat) != 0) {
        printf("Can't find file %s\n", fname);
        return -1;
    }

    fp = fopen(fname, "rb");
    if (fp == NULL) {
        printf("Can't open file %s: %s\n", fname, strerror(errno));
        return -1;
    }
    if (print_bitfile_header(fp, (char*) &part_name) == -1) {
        fclose(fp);
        return -1;
    }
/*    if (strcmp(part_name, active_board.fpga_part_number) != 0) {
        printf("Error: wrong bitfile destination device: %s, should be %s\n", part_name, active_board.fpga_part_number);
        fclose(fp);
        return -1;
    }*/
    if (check_boot_spi() == -1) {
        printf("Error: BootSector is invalid\n");
        fclose(fp);
        return -1;
    } else {
        printf("Boot sector OK\n");
    }

    printf("Verifying EEPROM sectors starting from 0x%X...\n", (unsigned int) start_address);
    printf("  |");
    fflush(stdout);
    i = start_address;
    while (!feof(fp)) {
        bytesread = fread(&file_buffer, 1, 8192, fp);
        bindex = 0;
        while (bindex < bytesread) {
            rdata = read_byte(bindex + i);
            mdata = file_buffer[bindex];
            if (mdata != rdata) {
                printf("\nError at 0x%X expected: 0x%X but read: 0x%X\n", i+bindex, mdata, rdata);
                return -1;
            }
            bindex = bindex + 1;
        }
        i = i + 8192;
        printf("V");
        fflush(stdout);
    }

    fclose(fp);
    printf("\nBoard configuration verified successfully\n");
    return 0;
}

int write_area_pci(mesa_board_t *board, char *fname, u32 start_address) {
    int bytesread, i;
    u32 eeprom_addr;
    char part_name[32];
    struct stat file_stat;
    FILE *fp;

    if (stat(fname, &file_stat) != 0) {
        printf("Can't find file %s\n", fname);
        return -1;
    }

    fp = fopen(fname, "rb");
    if (fp == NULL) {
        printf("Can't open file %s: %s\n", fname, strerror(errno));
        return -1;
    }
    if (print_bitfile_header(fp, (char*) &part_name) == -1) {
        fclose(fp);
        return -1;
    }
/*    if (strcmp(part_name, active_board.fpga_part_number) != 0) {
        printf("Error: wrong bitfile destination device: %s, should be %s\n", part_name, active_board.fpga_part_number);
        fclose(fp);
        return -1;
    }*/
    if (check_boot_spi() == -1) {
        write_boot_spi();
    } else {
        printf("Boot sector OK\n");
    }
    if (check_boot_spi() == -1) {
        printf("Failed to write valid boot sector\n");
        fclose(fp);
        return -1;
    }

    if (start_programming(board, start_address, file_stat.st_size) == -1) {
        fclose(fp);
        return -1;
    }
    printf("Programming EEPROM area starting from 0x%X...\n", (unsigned int) start_address);
    printf("  |");
    fflush(stdout);
    eeprom_addr = start_address;
    while (!feof(fp)) {
      bytesread = fread(&file_buffer, 1, 8192, fp);
      i = 0;
      printf("W");
      fflush(stdout);
      while (i < bytesread) {
        write_page(eeprom_addr, (char *) &file_buffer, i);
        i = i + PAGE_SIZE;
        eeprom_addr = eeprom_addr + PAGE_SIZE;
      }
    }
    printf("\nBoard configuration updated successfully\n");
    done_programming();
    return 0;
}

int open_pci_map(mesa_board_t *board) {
    switch (board->method) {
        case DEV_ACCESS_PCI: {
#ifdef __linux__
                int eno;

            map_size = board->dev->size[0];
            map_base_addr = board->dev->base_addr[0];

            if (seteuid(0) != 0) {
                errmsg(__func__, "need root privilges (or setuid root)");
                return -1;
            }
            memfd = open("/dev/mem", O_RDWR);
            eno = errno;
            seteuid(getuid());
            if (memfd < 0) {
                errmsg(__func__, "can't open /dev/mem: %s", strerror(eno));
                return -1;
            }

            mapped_ptr = mmap(0, map_size, PROT_READ | PROT_WRITE, MAP_SHARED, memfd, map_base_addr);
            if (mapped_ptr == MAP_FAILED) {
                mapped_ptr = NULL;
                errmsg(__func__, "can't map /dev/mem for mesa 5i25: %s", strerror(errno));
                close(memfd);
                return -1;
            }
#elif _WIN32
            iodll = LoadLibrary("winio32.dll");
            winio32_MapPhysToLin_ptr winio32_MapPhysToLin = NULL;

            phys.pvPhysAddress = ((DWORD64) board->dev->base_addr[0]) << 32;
            phys.dwPhysMemSizeInBytes = HM2_AREA_SIZE;
            map_size = HM2_AREA_SIZE;
            map_base_addr = board->dev->base_addr[0];

            if (iodll != NULL) {
                winio32_MapPhysToLin = (winio32_MapPhysToLin_ptr) GetProcAddress(iodll, "MapPhysToLin");
                if (winio32_MapPhysToLin != NULL)
                    mapped_ptr = winio32_MapPhysToLin(&phys);
            }
#endif
            read_cookie = &read_cookie_fpga;
            read_idrom = &read_idrom_spi;
            verify_area = &verify_area_pci;
            write_area = &write_area_pci;

            open_spi_access(DEV_ACCESS_PCI);
            board->fpga_part_number = "6slx9pq144";
            board->flash_id = read_idrom();
            board->user_address = calc_user_space(board->flash_id);
            prepare_boot_block(board->flash_id);

            pci_print_all_info(board, board->verbose);
            if (board->verbose != 0)
                pci_print_firmware_info();
        }
        break;

        case DEV_ACCESS_GPIO: {
            verify_area = &verify_area_pci;
            write_area = &write_area_pci;

            open_spi_access(DEV_ACCESS_GPIO);
            board->fpga_part_number = "6slx9pq144";
            board->user_address = 0x80000;

            printf("Board name: 6I25, access via XIO2001 bridge\n");
        }
        break;

        default:
        break;
    }
    return 0;
}

int close_pci_map(mesa_board_t *board) {
    switch (board->method) {
        case DEV_ACCESS_PCI: {
#ifdef __linux__
            munmap(mapped_ptr, map_size);
            close(memfd);
#elif _WIN32
            winio32_UnmapPhysicalMemory_ptr winio32_UnmapPhysicalMemory = NULL;

            if (iodll != NULL) {
                winio32_UnmapPhysicalMemory = (winio32_UnmapPhysicalMemory_ptr) GetProcAddress(iodll, "UnmapPhysicalMemory");
               if (winio32_UnmapPhysicalMemory != NULL)
                   winio32_UnmapPhysicalMemory(&phys);
            }
            FreeLibrary(iodll);
#endif
        }
        break;

        case DEV_ACCESS_GPIO: {
            close_spi_access(DEV_ACCESS_GPIO);
        }
        break;

        default:
        break;
    }

    return 0;
}
