
#ifdef __linux__
    #include <pci/pci.h>
#elif _WIN32
    #include "libpci/pci.h"
    #include <windows.h>
    #include <winsock2.h>
#endif
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>

#include "eeprom.h"
#include "devices.h"
#include "hostmot2.h"
#include "device_pci.h"
#include "device_eth.h"

u32 (*read_cookie)(void);
u8 (*read_idrom)(void);
int (*verify_area)(char *fname, u32 start_address);
int (*write_area)(mesa_board_t *board, char *fname, u32 start_address);

u8 file_buffer[SECTOR_SIZE] = {0, };

u8 boot_block[BOOT_BLOCK_SIZE] = {
    0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
    0xFF, 0xFF, 0xAA, 0x99, 0x55, 0x66, 0x31, 0xE1,
    0xFF, 0xFF, 0x32, 0x61, 0x00, 0x00, 0x32, 0x81,
    0x0B, 0x08, 0x32, 0xA1, 0x00, 0x00, 0x32, 0xC1,
    0x0B, 0x01, 0x30, 0xA1, 0x00, 0x00, 0x33, 0x01,
    0x21, 0x00, 0x32, 0x01, 0x00, 0x1F, 0x30, 0xA1,
    0x00, 0x0E, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
    0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00
};

// modify MSB of boot block jmp address to user area
void prepare_boot_block(u8 flash_id) {
    switch (flash_id) {
        case ID_EEPROM_1M:  boot_block[25] = 0x01; break;
        case ID_EEPROM_2M:  boot_block[25] = 0x02; break;
        case ID_EEPROM_4M:  boot_block[25] = 0x04; break;
        case ID_EEPROM_8M:  boot_block[25] = 0x08; break;
        case ID_EEPROM_16M: boot_block[25] = 0x10; break;
    }
}

u32 calc_user_space(u8 flash_id) {
    switch (flash_id) {
        case ID_EEPROM_1M:  return 0x10000; break;
        case ID_EEPROM_2M:  return 0x20000; break;
        case ID_EEPROM_4M:  return 0x40000; break;
        case ID_EEPROM_8M:  return 0x80000; break;
        case ID_EEPROM_16M: return 0x100000; break;
        default: return 0x80000; break;
    }
}

// prints info about bitfile and returns header length or -1 when error
int print_bitfile_header(FILE *fp, char *part_name) {
    u8 buff[256];
    int sleng;
    int bytesread, conflength;
    int ret = 0;

    printf("Checking file... ");
    bytesread = fread(&buff, 1, 14, fp);
    ret += bytesread;
    if (bytesread != 14) {
        if (feof(fp))
            printf("Error: Unexpected end of file\n");
        else
            printf("Error: IO error\n");
        return -1;
    }
    if ((buff[0] == 0) && (buff[1] == 9)) {
        printf("OK\n  File type: BIT file\n");
        if ((buff[11] == 0) && (buff[12] == 1) && (buff[13] == 0x61)) {
            bytesread = fread(&buff, 1, 2, fp);
            ret += bytesread;
            sleng = buff[0]*256 + buff[1];
            bytesread = fread(&buff, 1, sleng, fp);
            ret += bytesread;
            printf("  Design name: %s\n", buff);

            bytesread = fread(&buff, 1, 3, fp);
            ret += bytesread;
            sleng = buff[1]*256 + buff[2];
            bytesread = fread(&buff, 1, sleng, fp);
            ret += bytesread;
            printf("  Part name: %s\n", buff);
            strcpy(part_name, (char*) &buff);

            bytesread = fread(&buff, 1, 3, fp);
            ret += bytesread;
            sleng = buff[1]*256 + buff[2];
            bytesread = fread(&buff, 1, sleng, fp);
            ret += bytesread;
            printf("  Design date: %s\n", buff);

            bytesread = fread(&buff, 1, 3, fp);
            ret += bytesread;
            sleng = buff[1]*256 + buff[2];
            bytesread = fread(&buff, 1, sleng, fp);
            ret += bytesread;
            printf("  Design time: %s\n", buff);

            bytesread = fread(&buff, 1, 5, fp);
            ret += bytesread;
            conflength = buff[1]*16777216;
            conflength = conflength + buff[2]*65536;
            conflength = conflength + buff[3]*256;
            conflength = conflength + buff[4];
            printf("  Config Length: %d\n", conflength);
        }
        return ret;
    }
    if ((buff[0] == 0xFF) && (buff[1] == 0xFF) && (buff[2] == 0xFF) && (buff[3] == 0xFF)) {
        printf("Looks like a BIN file\n");
        return -1;
    }
    printf("Invalid bitfile header!\n");
    return -1;
}

//////////////////////////////////////

char *print_idrom_type(u8 idrom) {
    switch (idrom) {
        case ID_EEPROM_1M:  return "(size 1Mb)";
        case ID_EEPROM_2M:  return "(size 2Mb)";
        case ID_EEPROM_4M:  return "(size 4Mb)";
        case ID_EEPROM_8M:  return "(size 8Mb)";
        case ID_EEPROM_16M: return "(size 16Mb)";
        default:            return "(unknown size)";
    }
}

const char *hm2_get_general_function_name(u8 gtag) {
    switch (gtag) {
        case HM2_GTAG_WATCHDOG:        return "Watchdog";
        case HM2_GTAG_IOPORT:          return "IOPort";
        case HM2_GTAG_ENCODER:         return "Encoder";
        case HM2_GTAG_RESOLVER:        return "Resolver";
        case HM2_GTAG_STEPGEN:         return "StepGen";
        case HM2_GTAG_PWMGEN:          return "PWMGen";
        case HM2_GTAG_TRANSLATIONRAM:  return "TranslationRAM";
        case HM2_GTAG_TPPWM:           return "ThreePhasePWM";
        case HM2_GTAG_LED:             return "LED";
        case HM2_GTAG_MUXED_ENCODER:   return "Muxed Encoder";
        case HM2_GTAG_MUXED_ENCODER_SEL: return "Muxed Encoder Select";
        case HM2_GTAG_SMARTSERIAL:     return "Smart Serial Interface";
        case HM2_GTAG_BSPI:            return "Buffered SPI Interface";
        case HM2_GTAG_UART_RX:         return "UART Receive Channel";
        case HM2_GTAG_UART_TX:         return "UART Transmit Channel";
        default: {
            static char unknown[32];
            snprintf(unknown, 32, "(unknown-gtag-%d)", gtag);
            return unknown;
        }
    }
}

void errmsg(const char *funct, const char *fmt, ...) {
    va_list vp;

    va_start(vp, fmt);
    fprintf(stderr, "Error in %s(): ", funct);
    vfprintf(stderr, fmt, vp);
    fprintf(stderr, "\n");
    va_end(vp);
}

int open_device(mesa_board_t *board) {
    if (board == NULL) return -1;

    switch (board->method) {
        case DEV_ACCESS_PCI: {
            open_pci_map(board);
        }
        break;

        case DEV_ACCESS_GPIO: {
            open_pci_map(board);
        }
        break;

        case DEV_ACCESS_ETH: {
            open_eth_connection(board);
        }
        break;
    }
    return 0;
}

void close_device(mesa_board_t *board) {
    if (board == NULL) return;

    switch (board->method) {
        case DEV_ACCESS_PCI: {
            close_pci_map(board);
        }
        break;

        case DEV_ACCESS_GPIO: {
            close_pci_map(board);
        }
        break;

        case DEV_ACCESS_ETH: {
            close_eth_connection(board);
        }
        break;
    }
}
