
#ifdef __linux__
    #include <pci/pci.h>
    #include <sys/mman.h>
    #include <unistd.h>
#elif _WIN32
    #include "libpci/pci.h"
#endif
#include <fcntl.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>

#include "hostmot2.h"
#include "flash_spi.h"
#include "device_pci.h"

extern mesa_board_t *active_board;
extern u8 file_buffer[SECTOR_SIZE];
extern u8 boot_block[BOOT_BLOCK_SIZE];
extern void *mapped_ptr;

void (*set_cs_high)(void);
void (*send_byte)(u8 byte);
u8   (*recv_byte)(void);
void (*prefix)(void);
void (*suffix)(void);

static u16 GPIO_reg_val;

// functions to access flash via bridge GPIO

static void set_cs_high_gpio(void) {
    GPIO_reg_val = GPIO_reg_val | 0x2;
    pci_write_word(active_board->dev, XIO2001_GPIO_DATA_REG, GPIO_reg_val);
}

static void set_cs_low_gpio(void) {
    GPIO_reg_val = GPIO_reg_val & (~ 0x2);
    pci_write_word(active_board->dev, XIO2001_GPIO_DATA_REG, GPIO_reg_val);
}

static void set_din_high_gpio(void) {
    GPIO_reg_val = GPIO_reg_val | 0x8;
    pci_write_word(active_board->dev, XIO2001_GPIO_DATA_REG, GPIO_reg_val);
}

static void set_din_low_gpio(void) {
    GPIO_reg_val = GPIO_reg_val & (~0x8);
    pci_write_word(active_board->dev, XIO2001_GPIO_DATA_REG, GPIO_reg_val);
}

static void set_clock_high_gpio(void) {
    GPIO_reg_val = GPIO_reg_val | 0x10;
    pci_write_word(active_board->dev, XIO2001_GPIO_DATA_REG, GPIO_reg_val);
}

static void set_clock_low_gpio(void) {
    GPIO_reg_val = GPIO_reg_val & (~ 0x10);
    pci_write_word(active_board->dev, XIO2001_GPIO_DATA_REG, GPIO_reg_val);
}

static int get_bit_gpio(void) {
    u16 data;

    data = pci_read_word(active_board->dev, XIO2001_GPIO_DATA_REG);
    if (data & 0x4)
        return 1;
    else
        return 0;
}

static void prefix_gpio(void) {
    set_cs_high_gpio();
    set_din_low_gpio();
    set_clock_low_gpio();
    set_cs_low_gpio();
}

static void suffix_gpio(void) {
    set_cs_high_gpio();
    set_din_low_gpio();
}

static void init_gpio(void) {
    pci_write_word(active_board->dev, XIO2001_GPIO_ADDR_REG, 0x001B);
    pci_write_word(active_board->dev, XIO2001_SBAD_STAT_REG, 0x0000);
    GPIO_reg_val = 0x0003;
    pci_write_word(active_board->dev, XIO2001_GPIO_DATA_REG, GPIO_reg_val);
}

static void restore_gpio(void) {
    GPIO_reg_val = 0x0003;
    pci_write_word(active_board->dev, XIO2001_GPIO_DATA_REG, GPIO_reg_val);
    pci_write_word(active_board->dev, XIO2001_GPIO_ADDR_REG, 0x0000);
}

static void send_byte_gpio(u8 byte) {
    u32 mask = DATA_MASK;
    int i;

    for (i = 0; i < CMD_LEN; i++) {
        if ((mask & byte) == 0)
            set_din_low_gpio();
        else
            set_din_high_gpio();
        mask >>= 1;
        set_clock_high_gpio();
        set_clock_low_gpio();
    }
}

static u8 recv_byte_gpio(void) {
    u32 mask, data = 0;
    int i;

    mask = DATA_MASK;
    for (i = 0; i < DATA_LEN; i++) {
        if (get_bit_gpio() == 1)
            data |= mask;
        mask >>= 1;
        set_clock_high_gpio();
        set_clock_low_gpio();
    }
  return data;
}

// function to access flash via spi in HM2

static void set_cs_high_hm2(void) {
    PCI_WRITE_MEM32(HM2_SPI_CTRL_REG, 1);
}

static void set_cs_low(void) {
    PCI_WRITE_MEM32(HM2_SPI_CTRL_REG, 0);
}

static void prefix_hm2(void) {
    set_cs_low();
}

static void suffix_hm2(void) {
    set_cs_high_hm2();
}

static void wait_for_data(void) {
    u32 i = 0;
    volatile u8 data = 0;

    while ((data & DAV_MASK) == 0) {
        data = PCI_READ_MEM32(HM2_SPI_CTRL_REG);
        if (++i > 5000) {
            printf("%x timeout waiting for SPI data\n", data);
            return;
        }
    }
    return;
}

static void send_byte_hm2(u8 byte) {
    PCI_WRITE_MEM32(HM2_SPI_DATA_REG, byte);
    wait_for_data();
}

static u8 recv_byte_hm2(void) {
    PCI_WRITE_MEM32(HM2_SPI_DATA_REG, 0);
    wait_for_data();
    return PCI_READ_MEM32(HM2_SPI_DATA_REG);
}

// 

static void send_address(u32 addr) {
    send_byte((addr >> 16) & 0xFF);
    send_byte((addr >> 8) & 0xFF);
    send_byte(addr & 0xFF);
}

static void write_enable(void) {
  prefix();
  send_byte(CMD_WRITE_ENABLE);
  suffix();
}

static u8 read_status(void) {
    u8 ret;

    prefix();
    send_byte(CMD_READ_STATUS);
    ret = recv_byte();
    suffix();
    return ret;
}

static void wait_for_write(void) {
    u8 stat = read_status();

    while ((stat & WRITE_IN_PROGRESS_MASK) != 0) {
         stat = read_status();
    }
}

u8 read_idrom_spi(void) {
    u8 ret;

    prefix();
    send_byte(CMD_READ_IDROM);
    send_address(0); // three dummy bytes
    ret = recv_byte();
    suffix();
    return ret;
}

// exported functions

u8 read_byte(u32 addr) {
    u8 ret;

    prefix();
    send_byte(CMD_READ);
    send_address(addr);
    ret = recv_byte();
    suffix();
    return ret;
}

void erase_sector(u32 addr) {
    write_enable();
    prefix();
    send_byte(CMD_SECTOR_ERASE);
    send_address(addr);
    suffix();
    wait_for_write();
}

void write_page(u32 addr, char *buff, int buff_i) {
    int i;

    write_enable();
    prefix();
    send_byte(CMD_PAGE_WRITE);
    send_address(addr);// { note that add 0..7 should be 0}
    for (i = 0; i < PAGE_SIZE; i++) {
        send_byte(buff[buff_i + i]);
    }
    suffix();
    wait_for_write();
}

void done_programming(void) {
    set_cs_high();
    restore_gpio();
}

void open_spi_access(dev_access_method method) {
    switch (method) {
        case DEV_ACCESS_PCI: {
            set_cs_high = &set_cs_high_hm2;
            prefix = &prefix_hm2;
            suffix = &suffix_hm2;
            send_byte = &send_byte_hm2;
            recv_byte = &recv_byte_hm2;
        }
        break;

        case DEV_ACCESS_GPIO: {
            set_cs_high = &set_cs_high_gpio;
            prefix = &prefix_gpio;
            suffix = &suffix_gpio;
            send_byte = &send_byte_gpio;
            recv_byte = &recv_byte_gpio;

            init_gpio();
        }
        break;

        default:
        break;
    }
}

void close_spi_access(dev_access_method method) {
    switch (method) {
        case DEV_ACCESS_PCI: {
        }
        break;

        case DEV_ACCESS_GPIO: {
            restore_gpio();
        }
        break;

        default:
        break;
    }
}
