
#ifdef __linux__
    #include <pci/pci.h>
#elif _WIN32
    #include "libpci/pci.h"
#endif
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <getopt.h>
#include <ctype.h>

#include "devices.h"
#include "eeprom.h"
#include "hostmot2.h"
#include "device_pci.h"
#include "device_eth.h"
#include "lbp16.h"

static const char *unsupported_devs[] = {
    "5I20", "5I21", "5I22", "5I23",
    "7I43", "7I61",
    "3X20", NULL,
};

mesa_board_t boards[8];
int board_count = 0;
int bridges_count = 0;

mesa_board_t *active_board = NULL;

static int device_flag;
static int write_flag;
static int verify_flag;
static int info_flag;
static int ip_addr_flag;
static int scanpci_flag;
static int scaneth_flag;
static char eth_scan_addr[16];
static char bitfile_name[256];
static char device_name[16];
static char ip_addr[16] = "192.168.1.121";
static int set_ip_flag;
static char set_ip_addr[16];
static int recover_flag;
static int fallback_flag;
static int verbose_flag;
static int run_flag;

static struct option long_options[] = {
    {"scanpci", no_argument, &scanpci_flag, 1},
    {"scaneth", required_argument, 0, 'e'},
    {"recover", no_argument, &recover_flag, 1},
    {"fallback", no_argument, &fallback_flag, 1},
    {"device", required_argument, 0, 'd'},
    {"write", required_argument, 0, 'w'},
    {"verify", required_argument, 0, 'v'},
    {"verbose", no_argument, &verbose_flag, 1},
    {"info", required_argument, 0, 'i'},
    {"ip", required_argument, 0, 'a'},
    {"set", required_argument, 0, 's'},
    {"run", no_argument, &run_flag, 1},
    {0, 0, 0, 0}
};

int process_cmd_line(int argc, char *argv[]) {
    int c;

    while (1) {
        /* getopt_long stores the option index here. */
        int option_index = 0;

        c = getopt_long(argc, argv, "", long_options, &option_index);

        /* Detect the end of the options. */
        if (c == -1) break;
        switch (c) {
            case 0: {
                /* If this option set a flag, do nothing else now. */
                if (long_options[option_index].flag != 0) break;
                printf ("option %s", long_options[option_index].name);
                if (optarg)
                    printf (" with arg %s", optarg);
                printf ("\n");
            }
            break;

            case 'd': {
                int i;

                if (device_flag > 0) {
                    printf("Error: multiply --device option\n");
                    return -1;
                }
                strncpy(device_name, optarg, 16);
                for (i = 0; i < 16; i++)
                    device_name[i] = toupper(device_name[i]);

                device_flag++;
            }
            break;

            case 'a': {
                if (ip_addr_flag > 0) {
                    printf("Error: multiply --ip option\n");
                    return -1;
                }
                strncpy(ip_addr, optarg, 16);
                ip_addr_flag++;
            }
            break;

            case 'w': {
                if (write_flag > 0) {
                    printf("Error: multiply --device option\n");
                    return -1;
                }
                strncpy(bitfile_name, optarg, 255);
                write_flag++;
            }
            break;

            case 'v': {
                if (verify_flag > 0) {
                    printf("Error: multiply --verify option\n");
                    return -1;
                }
                strncpy(bitfile_name, optarg, 255);
                verify_flag++;
            }
            break;

            case 'i': {
                if (info_flag > 0) {
                    printf("Error: multiply --info option\n");
                    return -1;
                }
                strncpy(bitfile_name, optarg, 255);
                info_flag++;
            }
            break;

            case 'e': {
                if (scaneth_flag > 0) {
                    printf("Error: multiply --scaneth option\n");
                    return -1;
                }
                strncpy(eth_scan_addr, optarg, 32);
                scaneth_flag++;
            }
            break;

            case 's': {
                if (set_ip_flag > 0) {
                    printf("Error: multiply --set option\n");
                    return -1;
                }
                if (strncmp(optarg, "ip=", 3) == 0) {
                    strncpy(set_ip_addr, &optarg[3], 16);
                } else {
                    printf("unknown eeprom parameter %s\n", optarg);
                    return -1;
                }
                set_ip_flag++;
            }
            break;

            case '?':
                /* getopt_long already printed an error message. */
                return -1;
                break;

            default:
                abort();
        }
    }
    return 0;
}

void print_usage(void) {
   printf("Configuration program for mesanet 5I25/6I25/7I80 boards\n");
   printf("PCI device syntax:\n");
   printf("    mesaflash --scanpci\n");
   printf("    mesaflash --device device_name [--verbose] [[--recover] --write filename]\n");
   printf("    mesaflash --device device_name [--verbose] [--verify filename]\n");
   printf("ETH device syntax:\n");
   printf("    mesaflash --scaneth x.x.x.0\n");
   printf("    mesaflash --device device_name [--verbose] [--ip x.x.x.x] [--write | --verify filename]\n");
   printf("    mesaflash --device device_name [--ip x.x.x.x] --set param=value\n");
   printf("COMMON syntax:\n");
   printf("    mesaflash --info filename\n");
   printf("Options:\n");
   printf("  --scanpci     detect all pci(e) boards\n");
   printf("  --scaneth     detect all ethernet boards in given class C address (mask 255.255.255.0)\n");
   printf("  --device      select active device name (currently 5I25 | 6I25 | 7I80)\n");
   printf("  --write       write filename into flash (IMPORTANT! filename must be VALID FPGA configuration file)\n");
   printf("  --verify      verify if 'filename' is properly programmed in flash\n");
   printf("  --recover     write filename into flash using PCI bridge (currently only 6I25)\n");
   printf("  --ip          set ip address for connecting 7i80\n");
   printf("  --info        print info about configuration in filename\n");
   printf("  --verbose     print detailed information while running commands\n");
   printf("  --set         set 7I80 eeprom variable\n");
   printf("    ip=x.x.x.x  set board ethernet eeprom ip address\n");
}

int find_all_mesapci_devices(struct pci_access *pacc) {
    struct pci_dev *dev, *bridge = NULL;
    mesa_board_t *curr_board;

    // first find out if there is XIO2001 bridge
    for (dev = pacc->devices; dev != NULL; dev = dev->next) {
        pci_fill_info(dev, PCI_FILL_IDENT | PCI_FILL_BASES);     // first run - fill data struct 
        if ((dev->vendor_id == VENDORID_XIO2001) && (dev->device_id == DEVICEID_XIO2001)) {
            bridge = dev;
            bridges_count++;
        }
    }

    for (dev = pacc->devices; dev != NULL; dev = dev->next) {
        if ((dev->vendor_id == VENDORID_MESAPCI) && (dev->device_id == DEVICEID_MESA5I25)) {
            curr_board = &boards[board_count++];
            curr_board->type = DEV_5I25;
            curr_board->dev = dev;
            curr_board->bridge = bridge;
#ifdef _WIN32
            dev->size[0] = HM2_AREA_SIZE;    // on windows libpci doesn't report BAR length
#endif
        } else if ((dev->vendor_id == VENDORID_MESAPCI) && (dev->device_id == DEVICEID_MESA6I25)) {
            curr_board = &boards[board_count++];
            curr_board->type = DEV_6I25;
            curr_board->dev = dev;
            curr_board->bridge = bridge;
#ifdef _WIN32
            dev->size[0] = HM2_AREA_SIZE;    // on windows libpci doesn't report BAR length
#endif
        }
    }

    if (board_count == 0) {
        if ((bridges_count == 0)) {
            printf("No boards found.\n");
            return -1;
        } else if (bridges_count > 1) {
            printf("Error: (unimplemented): detected multiple XIO2001 bridges and no valid FPGAs.\n");
            return -1;
        } else {
            curr_board = &boards[board_count++];
            curr_board->dev = bridge;
            curr_board->bridge = bridge;
        }
    }
    return 0;
}

int main(int argc, char *argv[]) {
    struct pci_access *pacc;
    int libpci_inited = 0;
    u32 area_address;

    if (argc == 1) {
        print_usage();
        return 0;
    }

    if (process_cmd_line(argc, argv) == -1) {
        return -1;
    }

    if (info_flag == 1) {
        char part_name[32];
        struct stat file_stat;
        FILE *fp;

        if (stat(bitfile_name, &file_stat) != 0) {
            printf("Can't find file %s\n", bitfile_name);
            return -1;
        }

        fp = fopen(bitfile_name, "rb");
        print_bitfile_header(fp, (char*) &part_name);
        fclose(fp);
        return 0;
    }

    if (scanpci_flag == 1) {
        int i;

        pacc = pci_alloc();
        pci_init(pacc);            // inicjowanie biblioteki libpci
        pci_scan_bus(pacc);
        libpci_inited = 1;

        if (find_all_mesapci_devices(pacc) != 0) {
            return -1;
        }

        for (i = 0; i < board_count; i++) {
            pci_print_all_info(&boards[i], verbose_flag);
        }
        return 0;
    }
    else if (scaneth_flag == 1) {
        find_all_mesaeth_devices((char *) &eth_scan_addr);
        return 0;
    }

    if ((strncmp(device_name, "5I25", 4) == 0) && (strlen(device_name) == 4)) {
        int i;

        pacc = pci_alloc();
        pci_init(pacc);            // inicjowanie biblioteki libpci
        pci_scan_bus(pacc);
        libpci_inited = 1;

        if (find_all_mesapci_devices(pacc) != 0) {
            return -1;
        }

        for (i = 0; i < board_count; i++) {
            if (boards[i].type == DEV_5I25) {
                active_board = &boards[i];
                break;
            }
        }
        if (active_board == NULL) {
            printf("Error: Selected board 5I25 not found.\n");
            return 0;
        }
        active_board->method = DEV_ACCESS_PCI;
    } else if ((strncmp(device_name, "6I25", 4) == 0) && (strlen(device_name) == 4)) {
        int i;

        pacc = pci_alloc();
        pci_init(pacc);            // inicjowanie biblioteki libpci
        pci_scan_bus(pacc);
        libpci_inited = 1;

        if (find_all_mesapci_devices(pacc) != 0) {
            return -1;
        }

        for (i = 0; i < board_count; i++) {
            if (boards[i].type == DEV_6I25) {
                active_board = &boards[i];
                break;
            }
        }
        if (recover_flag == 0) {
            if (active_board == NULL) {
                printf("Error: Selected board 6I25 not found.\n");
                return 0;
            }
            active_board->method = DEV_ACCESS_PCI;
        } else if (recover_flag == 1) {
            for (i = 0; i < board_count; i++) {
                if (boards[i].bridge != NULL) {
                    active_board = &boards[i];
                    break;
                }
            }
            if (active_board == NULL) {
                printf("Error: Selected board 6I25 and recovery via GPIO but bridge XIO2001 not found.\n");
                return 0;
            }
            active_board->method = DEV_ACCESS_GPIO;
        }
    } else if ((strncmp(device_name, "7I80", 4) == 0) && (strlen(device_name) == 4))  {
        active_board = &boards[board_count++];
        active_board->type = DEV_7I80;
        strncpy(active_board->ip_addr, ip_addr, 16);
        active_board->method = DEV_ACCESS_ETH;
    } else {
        int i;

        for (i = 0; unsupported_devs[i] != NULL; i++) {
            if ((strncmp(device_name, unsupported_devs[i], 4) == 0) && (strlen(device_name) == 4)) {
                printf("Error: Device %s is unsupported\n", device_name);
                goto error;
            }
        }
        printf("Error: Device name %s is invalid\n", device_name);
        goto error;
    }

    active_board->verbose = verbose_flag;

    if (open_device(active_board) != 0) goto error;

    if (fallback_flag == 1) {
        area_address = FALLBACK_ADDRESS;
    } else {
        area_address = USER_ADDRESS;
    }

    if (verify_flag == 1) {
        verify_area(bitfile_name, area_address);
    } else if (write_flag == 1) {
        write_area(active_board, bitfile_name, area_address);
    } else if (set_ip_flag == 1) {
        lbp16_send_write_ip_address((char *) &set_ip_addr);
    } else if (run_flag == 1) {
        run();
    }

    close_device(active_board);

error:
    if (libpci_inited == 1) {
        pci_cleanup(pacc);
    }
    return 0;
}
